import std/[logging, json, options, os, tables, times]
import timeutils, uuids, zero_functional

import ./jsonutils, ./models

type
  DbRoot = ref object
    albums: TableRef[UUID, Album]
    artists: TableRef[UUID, Artist]
    bookmarks: TableRef[UUID, Bookmark]
    mediaFiles: TableRef[UUID, MediaFile]
    playlists: TableRef[UUID, Playlist]
    tags: TableRef[string, Option[string]]

    albumsToMediaFiles: TableRef[UUID, seq[UUID]]
    artistsToMediaFiles: TableRef[UUID, seq[UUID]]
    artistsAndAlbums: seq[tuple[artistId, albumId: UUID]]
    playlistsToMediaFiles: TableRef[UUID, seq[UUID]]
    tagsAndMediaFiles: seq[tuple[tagName: string, mediaFileId: UUID]]

    mediaFileHashToId: TableRef[string, UUID]

  WdiwtltDb* = ref object
    jsonFilePath: string
    root: DbRoot

proc debug*(db: WdiwtltDb): string =
  "\p\talbums: " & $db.root.albums &
  "\p\tartists: " & $db.root.artists &
  "\p\tbookmarks: " & $db.root.bookmarks &
  "\p\tmediaFiles: " & $db.root.mediaFiles &
  "\p\tplaylists: " & $db.root.playlists &
  "\p\ttags: " & $db.root.tags &
  "\p\talbumsToMediaFiles: " & $db.root.albumsToMediaFiles &
  "\p\tartistsToMediaFiles: " & $db.root.artistsToMediaFiles &
  "\p\tartistsAndAlbums: " & $db.root.artistsAndAlbums &
  "\p\tplaylistsToMediaFiles: " & $db.root.playlistsToMediaFiles &
  "\p\ttagsAndMediaFiles: " & $db.root.tagsAndMediaFiles &
  "\p\tmediaFileHashToId: " & $db.root.mediaFileHashToId & "\p"

## ========================================================================= ##
##                            API Contract                                   ##
## ========================================================================= ##
proc initDbRoot(): DbRoot;
proc initDb*(path: string): WdiwtltDb;
proc loadDb*(path: string): WdiwtltDb;
proc persist*(db: WdiwtltDb): void;

## Albums
## --------------------
proc add*(db: WdiwtltDb, a: Album);
proc remove*(db: WdiwtltDb, a: Album);
proc update*(db: WdiwtltDb, a: Album);

proc findAlbumsByArtist*(db: WdiwtltDb, a: Artist): seq[Album];
proc findAlbumsByName*(db: WdiwtltDb, name: string): seq[Album];

proc add*(db: WdiwtltDb, a: Album, mf: MediaFile);
proc remove*(db: WdiwtltDb, a: Album, mf: MediaFile);

## Artists
## --------------------
proc add*(db: WdiwtltDb, a: Artist);
proc remove*(db: WdiwtltDb, a: Artist);
proc update*(db: WdiwtltDb, a: Artist);

proc findArtistsByAlbum*(db: WdiwtltDb, a: Album): seq[Artist];
proc findArtistsByName*(db: WdiwtltDb, name: string): seq[Artist];

proc add*(db: WdiwtltDb, a: Artist, mf: MediaFile);
proc remove*(db: WdiwtltDb, a: Artist, mf: MediaFile);

proc isAssociated*(db: WdiwtltDb, artist: Artist, album: Album): bool;
proc associate*(db: WdiwtltDb, artist: Artist, album: Album);
proc disassociate*(db: WdiwtltDb, artist: Artist, album: Album);

## Bookmarks
## --------------------
proc add*(db: WdiwtltDb, b: Bookmark);
proc update*(db: WdiwtltDb, b: Bookmark);
proc remove*(db: WdiwtltDb, b: Bookmark);

## Media Files
## --------------------
proc add*(db: WdiwtltDb, mf: MediaFile);
proc remove*(db: WdiwtltDb, mf: MediaFile);
proc update*(db: WdiwtltDb, mf: MediaFile);
proc findMediaFilesByAlbum*(db: WdiwtltDb, a: Album): seq[MediaFile];
proc findMediaFilesByArtist*(db: WdiwtltDb, a: Artist): seq[MediaFile];
proc findMediaFileByHash*(db: WdiwtltDb, hash: string): Option[MediaFile];

## Playlists
## --------------------
proc add*(db: WdiwtltDb, p: Playlist);
proc remove*(db: WdiwtltDb, p: Playlist);
proc update*(db: WdiwtltDb, p: Playlist);

## Tags
## --------------------
proc add*(db: WdiwtltDb, t: Tag);
proc remove*(db: WdiwtltDb, t: Tag);
proc update*(db: WdiwtltDb, t: Tag);

## Housekeeping
## --------------------
proc pruneStalePlaylists*(db: WdiwtltDb, ts: DateTime);
proc pruneStaleBookmarks*(db: WdiwtltDb, ts: DateTime);
proc removeEmptyAlbums*(db: WdiwtltDb): void;
proc removeEmptyArtists*(db: WdiwtltDb): void;
proc removeEmptyPlaylists*(db: WdiwtltDb): void;

## ========================================================================= ##
##                            API Implementation                             ##
## ========================================================================= ##

## To JSON
## --------------------
proc `%`(dt: DateTime): JsonNode = %(dt.formatIso8601)
proc `%`(table: TableRef): JsonNode =
  result = newJObject()
  for k, v in table.pairs: result[$k] = %v

proc `%`(tup: tuple[albumId, mediaFileId: UUID]): JsonNode =
  result = newJObject()
  result["albumId"] = %($tup.albumId)
  result["mediaFileId"] = %($tup.mediaFileId)

proc `%`(tup: tuple[artistId, mediaFileId: UUID]): JsonNode =
  result = newJObject()
  result["artistId"] = %($tup.artistId)
  result["mediaFileId"] = %($tup.mediaFileId)

proc `%`(tup: tuple[artistId, albumId: UUID]): JsonNode =
  result = newJObject()
  result["artistId"] = %($tup.artistId)
  result["albumId"] = %($tup.albumId)

proc `%`(tup: tuple[tagName: string, mediaFileId: UUID]): JsonNode =
  result = newJObject()
  result["tagName"] = %($tup.tagName)
  result["mediaFileId"] = %($tup.mediaFileId)

proc `%`(root: DbRoot): JsonNode =
  %*{
    "albums": root.albums,
    "artists": root.artists,
    "bookmarks": root.bookmarks,
    "mediaFiles": root.mediaFiles,
    "playlists": root.playlists,
    "tags": root.tags,
    "albumsToMediaFiles": root.albumsToMediaFiles,
    "artistsToMediaFiles": root.artistsToMediaFiles,
    "artistsAndAlbums": root.artistsAndAlbums,
    "playlistsToMediaFiles": root.playlistsToMediaFiles,
    "tagsAndMediaFiles": root.tagsAndMediaFiles,
    "mediaFileHashToId": root.mediaFileHashToId
  }

## From JSON
## --------------------
proc parseAlbumsTable(n: JsonNode): TableRef[UUID, Album] =
  result = newTable[UUID, Album]()
  for strId in n.keys: result[parseUUID(strId)] = parseAlbum(n[strId])

proc parseArtistsTable(n: JsonNode): TableRef[UUID, Artist] =
  result = newTable[UUID, Artist]()
  for strId in n.keys: result[parseUUID(strId)] = parseArtist(n[strId])

proc parseBookmarksTable(n: JsonNode): TableRef[UUID, Bookmark] =
  result = newTable[UUID, Bookmark]()
  for strId in n.keys: result[parseUUID(strId)] = parseBookmark(n[strId])

proc parseMediaFilesTable(n: JsonNode): TableRef[UUID, MediaFile] =
  result = newTable[UUID, MediaFile]()
  for strId in n.keys: result[parseUUID(strId)] = parseMediaFile(n[strId])

proc parsePlaylistsTable(n: JsonNode): TableRef[UUID, Playlist] =
  result = newTable[UUID, Playlist]()
  for strId in n.keys: result[parseUUID(strId)] = parsePlaylist(n[strId])

proc parseTagsTable(n: JsonNode): TableRef[string, Option[string]] =
  result = newTable[string, Option[string]]()
  for tagName in n.keys:
    if n[tagName].getStr("").len > 0: result[tagName] = some(n[tagName].getStr)
    else: result[tagName] = none[string]()

proc parseMediaFilesAssociationTable(n: JsonNode): TableRef[UUID, seq[UUID]] =
  result = newTable[UUID, seq[UUID]]()
  for strId in n.keys:
    result[parseUUID(strId)] = n[strId].getElems --> map(parseUUID(it.getStr))

proc parseArtistsAndAlbumsList(n: JsonNode):
    seq[tuple[artistId, albumId: UUID]] =
  n.getElems --> map((it.parseUUID("artistId"), it.parseUUID("albumId")))

proc parseTagsAndMediaFilesList(n: JsonNode):
  seq[tuple[tagName: string, mediaFileId: UUID]] =
  n.getElems --> map((it.getOrFail("tagName").getStr, it.parseUUID("mediaFileId")))

proc parseMediaFileHashToIdTable(n: JsonNode): TableRef[string, UUID] =
  result = newTable[string, UUID]()
  for hash in n.keys: result[hash] = parseUUID(n[hash].getStr)

proc parseDbRoot(n: JsonNode): DbRoot =
  result = DbRoot(
    albums: n.getOrFail("albums").parseAlbumsTable,
    artists: n.getOrFail("artists").parseArtistsTable,
    bookmarks: n.getOrFail("bookmarks").parseBookmarksTable,
    mediaFiles: n.getOrFail("mediaFiles").parseMediaFilesTable,
    playlists: n.getOrFail("playlists").parsePlaylistsTable,
    tags: n.getOrFail("tags").parseTagsTable,

    albumsToMediaFiles:
      n.getOrFail("albumsToMediaFiles").parseMediaFilesAssociationTable,
    artistsToMediaFiles:
      n.getOrFail("artistsToMediaFiles").parseMediaFilesAssociationTable,
    artistsAndAlbums:
      n.getOrFail("artistsAndAlbums").parseArtistsAndAlbumsList,
    playlistsToMediaFiles:
      n.getOrFail("playlistsToMediaFiles").parseMediaFilesAssociationTable,
    tagsAndMediaFiles:
      n.getOrFail("tagsAndMediaFiles").parseTagsAndMediaFilesList,
    mediaFileHashToId:
      n.getOrFail("mediaFileHashToId").parseMediaFileHashToIdTable
  )


## Internals
## --------------------
proc initDbRoot(): DbRoot =
  DbRoot(
    albums: newTable[UUID, Album](),
    artists: newTable[UUID, Artist](),
    bookmarks: newTable[UUID, Bookmark](),
    mediaFiles: newTable[UUID, MediaFile](),
    playlists: newTable[UUID, Playlist](),
    tags: newTable[string, Option[string]](),
    albumsToMediaFiles: newTable[UUID, seq[UUID]](),
    artistsToMediaFiles: newTable[UUID, seq[UUID]](),
    artistsAndAlbums: @[],
    playlistsToMediaFiles: newTable[UUID, seq[UUID]](),
    tagsAndMediaFiles: @[],
    mediaFileHashToId: newTable[string, UUID]())

proc initDb*(path: string): WdiwtltDb =
  WdiwtltDb(
    jsonFilePath: path,
    root: initDbRoot())

proc loadDb*(path: string): WdiwtltDb =
  if not fileExists(path):
    raise newException(Exception, "Unable to open database file '" & path & "'")

  debug "loaded DB"
  result = WdiwtltDb(
    jsonFilePath: path,
    root: parseDbRoot(parseJson(path.readFile)))
  debug result.debug

proc persist*(db: WdiwtltDb): void =
  db.jsonFilePath.writeFile($(%db.root))

## Albums
## --------------------
proc add*(db: WdiwtltDb, a: Album) = db.root.albums[a.id] = a

proc remove*(db: WdiwtltDb, a: Album) =
  db.root.albums.del(a.id)
  db.root.albumsToMediaFiles.del(a.id)
  db.root.artistsAndAlbums =
    db.root.artistsAndAlbums --> filter(it.albumId != a.id)

proc update*(db: WdiwtltDb, a: Album) = db.add(a)

proc findAlbumsByArtist*(db: WdiwtltDb, a: Artist): seq[Album] =
  db.root.artistsAndAlbums -->
    filter(it.artistId == a.id and db.root.artists.contains(a.id)).
    map(db.root.albums[it.albumId])

proc findAlbumsByName*(db: WdiwtltDb, name: string): seq[Album] =
  result = @[]
  for a in db.root.albums.values:
    if a.name == name: result.add(a)

proc add*(db: WdiwtltDb, a: Album, mf: MediaFile) =
  if not db.root.albumsToMediaFiles.contains(a.id):
    db.root.albumsToMediaFiles[a.id] = @[mf.id]
  else:
    let idsInAlbum = db.root.albumsToMediaFiles[a.id]
    if not idsInAlbum.contains(mf.id):
      db.root.albumsToMediaFiles[a.id].add(mf.id)

proc remove*(db: WdiwtltDb, a: Album, mf: MediaFile) =
  if db.root.albumsToMediaFiles.contains(a.id):
    db.root.albumsToMediaFiles[a.id] =
      db.root.albumsToMediaFiles[a.id] --> filter(it != mf.id)

## Artists
## --------------------
proc add*(db: WdiwtltDb, a: Artist) = db.root.artists[a.id] = a

proc remove*(db: WdiwtltDb, a: Artist) =
  db.root.artists.del(a.id)
  db.root.artistsToMediaFiles.del(a.id)
  db.root.artistsAndAlbums =
    db.root.artistsAndAlbums --> filter(it.artistId != a.id)

proc update*(db: WdiwtltDb, a: Artist) = db.add(a)

proc findArtistsByAlbum*(db: WdiwtltDb, a: Album): seq[Artist] =
  db.root.artistsAndAlbums -->
    filter(it.albumId == a.id and db.root.albums.contains(a.id)).
    map(db.root.artists[it.artistId])

proc findArtistsByName*(db: WdiwtltDb, name: string): seq[Artist] =
  result = @[]
  for a in db.root.artists.values:
    if a.name == name: result.add(a)

proc add*(db: WdiwtltDb, a: Artist, mf: MediaFile) =
  if not db.root.artistsToMediaFiles.contains(a.id):
    db.root.artistsToMediaFiles[a.id] = @[mf.id]
  else:
    let idsInArtist = db.root.artistsToMediaFiles[a.id]
    if not idsInArtist.contains(mf.id):
      db.root.artistsToMediaFiles[a.id].add(mf.id)

proc remove*(db: WdiwtltDb, a: Artist, mf: MediaFile) =
  if db.root.artistsToMediaFiles.contains(a.id):
    db.root.artistsToMediaFiles[a.id] =
      db.root.artistsToMediaFiles[a.id] --> filter(it != mf.id)

proc isAssociated*(
    db: WdiwtltDb,
    artist: Artist,
    album: Album
  ): bool =
  let matching = db.root.artistsAndAlbums -->
    filter(it.artistId == artist.id and it.albumId == album.id)
  return matching.len > 0

proc associate*(db: WdiwtltDb, artist: Artist, album: Album) =
  if not db.isAssociated(artist, album):
    db.root.artistsAndAlbums.add((artist.id, album.id))

proc disassociate*(db: WdiwtltDb, artist: Artist, album: Album) =
  db.root.artistsAndAlbums = db.root.artistsAndAlbums -->
    filter(it.albumId != album.id or it.artistId != artist.id)

## Bookmarks
## --------------------
proc add*(db: WdiwtltDb, b: Bookmark) = db.root.bookmarks[b.id] = b
proc remove*(db: WdiwtltDb, b: Bookmark) = db.root.bookmarks.del(b.id)
proc update*(db: WdiwtltDb, b: Bookmark) = db.add(b)

## Media Files
## --------------------
proc add*(db: WdiwtltDb, mf: MediaFile) =
  db.root.mediaFiles[mf.id] = mf
  db.root.mediaFileHashToId[mf.fileHash] = mf.id

proc remove*(db: WdiwtltDb, mf: MediaFile) =
  db.root.mediaFiles.del(mf.id)

  for albumId, mfIds in db.root.albumsToMediaFiles.pairs:
    if mfIds.contains(mf.id):
      db.remove(db.root.albums[albumId], mf)

  db.root.mediaFileHashToId.del(mf.fileHash)

proc update*(db: WdiwtltDb, mf: MediaFile) = db.add(mf)

proc findMediaFilesByAlbum*(db: WdiwtltDb, a: Album): seq[MediaFile] =
  result = @[]
  if db.root.albumsToMediaFiles.contains(a.id):
    return db.root.albumsToMediaFiles[a.id] -->
      filter(db.root.mediaFiles.contains(it)).
      map(db.root.mediaFiles[it])

proc findMediaFilesByArtist*(db: WdiwtltDb, a: Artist): seq[MediaFile] =
  result = @[]
  if db.root.artistsToMediaFiles.contains(a.id):
    return db.root.artistsToMediaFiles[a.id] -->
      filter(db.root.mediaFiles.contains(it)).
      map(db.root.mediaFiles[it])

proc findMediaFileByHash*(db: WdiwtltDb, hash: string): Option[MediaFile] =
  if db.root.mediaFileHashToId.contains(hash):
    let mfId = db.root.mediaFileHashToId[hash]
    if db.root.mediaFiles.contains(mfId):
      return some(db.root.mediaFiles[mfId])

  return none[MediaFile]()

## Playlists
## --------------------
proc add*(db: WdiwtltDb, p: Playlist) = db.root.playlists[p.id] = p

proc remove*(db: WdiwtltDb, p: Playlist) =
  db.root.playlists.del(p.id)
  db.root.playlistsToMediaFiles.del(p.id)
  for b in db.root.bookmarks.values:
    if b.playlistId == p.id: db.remove(b)

proc update*(db: WdiwtltDb, p: Playlist) = db.add(p)

## Tags
## --------------------
proc add*(db: WdiwtltDb, t: Tag) = db.root.tags[t.name] = t.description

proc remove*(db: WdiwtltDb, t: Tag) =
  db.root.tags.del(t.name)
  db.root.tagsAndMediaFiles = db.root.tagsAndMediaFiles -->
    filter(it.tagName != t.name)

proc update*(db: WdiwtltDb, t: Tag) = db.add(t)

## Housekeeping
## --------------------
proc pruneStalePlaylists*(db: WdiwtltDb, ts: DateTime) =
  db.root.playlists.values -->
    filter(it.lastUsed > ts).foreach(db.remove(it))

proc pruneStaleBookmarks*(db: WdiwtltDb, ts: DateTime) =
  db.root.bookmarks.values -->
    filter(it.lastUsed > ts).foreach(db.remove(it))

proc removeEmptyAlbums*(db: WdiwtltDb): void =
  let emptyAlbumIds = db.root.albumsToMediaFiles.pairs -->
    filter(it[1].len == 0).map(it[0])

  for i in emptyAlbumIds:
    db.root.albums.del(i)
    db.root.albumsToMediaFiles.del(i)
    db.root.artistsAndAlbums = db.root.artistsAndAlbums -->
      filter(it.albumId != i)

proc removeEmptyArtists*(db: WdiwtltDb): void =
  let emptyArtistIds = db.root.artistsToMediaFiles.pairs -->
    filter(it[1].len == 0).map(it[0])

  for i in emptyArtistIds:
    db.root.artists.del(i)
    db.root.artistsToMediaFiles.del(i)
    db.root.artistsAndAlbums = db.root.artistsAndAlbums -->
      filter(it.artistId != i)


proc removeEmptyPlaylists*(db: WdiwtltDb): void =
  let emptyPlaylistIds = db.root.playlistsToMediaFiles.pairs -->
    filter(it[1].len == 0).map(it[0])

  for i in emptyPlaylistIds:
    db.root.playlists.del(i)
    db.root.playlistsToMediaFiles.del(i)
    let emptyBookmarkIds = db.root.bookmarks.pairs -->
      filter(it[1].playlistId == i).map(it[0])
    for j in emptyBookmarkIds: db.root.bookmarks.del(j)
